""" Constants """
import json
OPTION_PRECEDENCE = ['cli', 'env', 'ini', 'dft']

OPTION_TYPE_MAP = {
    # 'protocol': str,
    'debug': str,
    'local_strategy': bool,
    'no_config_cache': bool,
    'no_config_sync': bool,
    'murano_host': str,
    'murano_id': str,
    'no_filesystem': bool,
    'config_io_file': str,
    'ini_file': str,
    'murano_token': str,
    'watchlist': list,
    'http_timeout': int,
    'certfile': str,
    'pkeyfile': str,
    'murano_cacert': str,
    'murano_port': int
}

OPTION_NAME_MAPPER = {
    'env': lambda d, x: d.get('EDGED_{}'.format(x.upper()), None),
    'cli': lambda d, x: d.get('--{}'.format(x.replace('_', '-')), None),
    'dft': lambda d, x: d.get(x, None),
    'ini': lambda d, x: d.get(x)
}

def _ini_get(I, string):
    try:
        s, o = string.split('.')
        if s in I.sections():
            return I.get(s, o)
        else:
            return None
    except:
        return None

DEFAULTS = {
    'protocol': 'http',
    'debug': 'CRITICAL',
    'local_strategy': False,
    'no_config_cache': False,
    'no_config_sync': False,
    'no_filesystem': False,
    'watchlist': ['config_io'],
    'http_timeout': 60*5*1000,
    'ini_file': 'default.ini'
}



"""
Module for testing data types
"""
import logging


def iam_object(**kwargs):
    return kwargs


def iam_list(**kwargs):
    return kwargs.values()


def iam_tuple(**kwargs):
    return zip(kwargs.keys(), kwargs.values())[0]


def iam_bool(**kwargs):
    key = kwargs.keys()[0]
    return key == kwargs[key]


# pylint: disable=C0103
""" Sample ExoEdge application

This module provides utilities for simulating data to enable a quick-start
experience with ExoEdge and the IIoT solution ExoSense.
"""
from __future__ import print_function
import time
import random
import threading
import logging
import math
import json
import re


def fourteen():
    """ Return 14 """
    return 14


def current_time():
    """ Return the current timestamp """
    return time.time()


def echo(value=''):
    """ Echo a value into a string

    Parameters:
    value:          The value to be echoed
    """
    return str(value)


def strip_non_numeric(value=''):
    """ Strip out non-numeric characters from string

    Parameters:
    value:          The value to be stripped to numeric values
    """
    return int(re.sub('[^0-9]', '', value))


def sin_wave(period=60, amplitude=1, offset=0, precision=2):
    """ Generate a sin wave from the current time

    Parameters:
    period:         The period, in seconds, of the sin wave
    amplitude:      The amplitude of the sin wave
    offset:         The vertical offset of the sin wave
    """
    ratio = (time.time() % period) / period
    return round(math.sin(2 * math.pi * ratio) * amplitude + offset, precision)


def cos_wave(period=60, amplitude=1, offset=0, precision=2):
    """ Generate a sin wave from the current time

    Parameters:
    period:         The period, in seconds, of the cos wave
    amplitude:      The amplitude of the cos wave
    offset:         The vertical offset of the cos wave
    """
    ratio = (time.time() % period) / period
    return round(math.cos(2 * math.pi * ratio) * amplitude + offset, precision)


def location(latitude=None, longitude=None, period=60, radius=0.1, precision=6):
    """ Generate location data from the current time and a starting location

    Moves in a circle every `period` seconds. Path is `radius` decimal degrees
    from the center point defined by `latitude` and `longitude`.

    Parameters:
    latitude:       The latitude of the center point, in decimal degrees
    longitude:      The longitude of the center point, in decimal degrees
    period:         The period, in seconds, it takes to traverse the path
    radius:         The radius, in decimal degrees, of the path from the center
    """
    lat = sin_wave(period=period,
                   amplitude=radius,
                   offset=latitude,
                   precision=precision)
    lng = cos_wave(period=period,
                   amplitude=radius,
                   offset=longitude,
                   precision=precision)
    return json.dumps({'lat': lat, 'lng': lng})


def random_integer(lower=0, upper=10):
    """ Get a random integer between two values

    Parameters:
    lower:          The lower bound of the random number
    upper:          The upper bound of the random number
    """
    return random.randint(lower, upper)


def random_sleep_1(_fn=None, event=None, lower=1, upper=10):
    """ Start AsyncSim thread.

    The two parameters `_fn` and `event` are protected and overwritten
    by exoedge.channel.Channel

    Parameters:
    lower:          The lower bound of the sleep time, in seconds
    upper:          The upper bound of the sleep time, in seconds
    """
    AsyncSim(_fn, event, lower, upper).start()


class AsyncSim(threading.Thread):
    """ Minimal class to simulate a threaded source """
    def __init__(self, f, e, lower, upper):
        """
        Initialized by async(). For this example, setDaemon is used so that,
        if the channel that created this thread is killed, the AsyncSim thread
        will also die.
        """
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self._fn = f
        self.event = e
        self.lower = lower
        self.upper = upper

    def run(self):
        """ Sleep for a random number of seconds

        After the thread has slept, set an event passed by its owner Channel.
        """
        logging.info('starting')
        while True:
            r = random.randint(self.lower, self.upper)
            logging.info('sleeping %ss', r)
            time.sleep(r)
            self._fn(current_time())
            self.event.set()


def random_sleep_2(lower=1, upper=10):
    """ Sleep for a random number of seconds

    Distinct from random_sleep_1 in that there is no additional thread
    required to get a value into the Channel. Instead, it merely returns when
    the sleep is done.

    Parameters:
    lower:          The lower bound of the sleep time, in seconds
    upper:          The upper bound of the sleep time, in seconds
    """
    r = random.randint(lower, upper)
    logging.info('sleeping %ss', r)
    time.sleep(r)
    return current_time()



import unittest
from exoedge.exo_edge import ExoEdge, NoLocalConfigIO

class TestExoEdge(unittest.TestCase):
    def test_no_local_config(self):
        with self.assertRaises(NoLocalConfigIO):
            ExoEdge('', **{'local_strategy': True,
                           'config_io_file': '/file/dne'}).setup()
    
    def test_bool_init_logic(self):
        E = ExoEdge('', **{'local_strategy': True,
                           'config_io_file': '/file/dne',
                           'no_config_cache': True,
                           'no_config_sync': None})
        self.assertFalse(E.cache_config_io)
        self.assertTrue(E.config_io_sync)
        self.assertIsNotNone(E.config_io_watcher)

    def test_default_config_file(self):
        E = ExoEdge('', **{'murano_id': 'MYMURANOID'})
        self.assertEquals(E.config_io_file, 'MYMURANOID.json')


def main():
    unittest.main()

if __name__ == "__main__":
    main()



# pylint: disable=C0103,R0902
"""
Module to generate data from arbitrary source modules as defined in config_io.
"""
import os
import sys
import time
import logging
import json
import threading
import six
if six.PY2:
    import Queue as queue
elif six.PY3:
    import queue
from murano_client.client import StoppableThread


path = os.path.dirname(
    os.path.realpath(__file__)
).rsplit('/', 1)[0] + '/exoedge/sources'
sys.path.append(path)

LOG = logging.getLogger('EXOEDGE.' + __name__)


class Channel(StoppableThread, object):
    """
    A Channel is connected to a Source, which provides the Channel
    with data to send in data_in.

    This class provides methods to get new data from a source specified in
    config_io.
    """
    def __init__(self, **kwargs):
        """
        Channel initialized by config_io.ConfigIO._add_channel()

        Parameters:
        name:               The key used for the channel object in config_io
        display_name:       The human-readable display name of the Channel
        description:        The description of the Channel
        properties:         <<<General metadata store>>>
                            <<<TODO>>>
        protocol_config:    The configuration for the source protocol
        offset:             The `b` parameter in `y = mx + b` to be applied to
                            a value before being put in the Channel's queue
        multiplier:         The `m` parameter in `y = mx + b` to be applied to
                            a value before being put in the Channel's queue
        sample_rate:        The period in milliseconds between source samples
        report_rate:        The period in milliseconds between reports to
                            Murano
        report_on_change:   Boolean to only send value if the value has changed
                            since last sample
        mode:               The mode with which the Source is run.
                            poll | async | async2
        down_sample:        Method by which to down sample values in Channel
                            queue
                            MAX | MIN | SUM | AVG | ACT
        async_timeout:      The timeout in seconds for waiting on the event
                            set by an `async`-mode source.
        app_specific_config The reference object to the source from which data
                            will be received
        """
        StoppableThread.__init__(self, name=kwargs.get('name'))
        # TOP LEVEL KEYS
        self.name = kwargs.get('name')
        self.display_name = kwargs.get('display_name')
        self.description = kwargs.get('description')
        self.properties = kwargs.get('properties')
        self.protocol_config = kwargs.get('protocol_config', {})

        # PROTOCOL_CONFIG KEYS
        self.application = self.protocol_config.get('application')
        self.interface = self.protocol_config.get('interface')
        self.offset = self.protocol_config.get('offset', 0)
        self.multiplier = self.protocol_config.get('multiplier', 1)   # "gain"
        self.sample_rate = self.protocol_config.get('sample_rate')
        self.report_rate = self.protocol_config.get('report_rate')
        self.report_on_change = self.protocol_config.get('report_on_change')
        self.mode = self.protocol_config.get('mode', 'poll')
        self.down_sample = self.protocol_config.get('down_sample', 'ACT')

        self.app_specific_config = self.protocol_config.get('app_specific_config', {})
        self.positionals = self.app_specific_config.get('positionals', tuple())
        self.parameters = self.app_specific_config.get('parameters', {})
        self.__script = self.app_specific_config.get('__script', '')

        # DERIVED
        self.downsampler = DownSampler(self.down_sample)
        self.async_timeout = kwargs.get('async_timeout', 1.0)
        self.last_value = None
        self.last_report = 0
        self.q_out = queue.Queue() if self.down_sample else queue.LifoQueue()

        self.e_emission = threading.Event()
        if self.mode == 'async':
            self.e_async = threading.Event()
            self.parameters.update({
                '_fn': self.put_data,
                'event': self.e_async
            })

    def get_data(self):
        """ Get value from Source.

        Import Source module and run specified function, passing in
        Channel.source.parameters as keyword arguments.
        """
        if not self.__script:
            val = getattr(__import__(self.app_specific_config['module']),
                          self.app_specific_config['function'])(*self.positionals, **self.parameters)
            return val
        else:
            try:
                exec(self.__script)
                return val
            except (SyntaxError, NameError):
                raise Exception('__script must define `val`')
        return val

    def handle_getting_of_data(self):
        """ Call get_data() while handling different Source modes.

        TODO: rename this method
        """
        if self.mode == 'poll':
            # Typical behavior. Poll a function at sample_rate
            # and send if report_on_change is false or the value
            # has changed.
            _new_val = self.get_data()
            if self.report_on_change and _new_val == self.last_value:
                # only report_on_change, even if report_rate passes
                self._sleep()
            else:
                self.try_emit()
                self.put_data(_new_val)
                self.last_value = _new_val
                self._sleep()

        elif self.mode == 'async':
            # Used if the function defined has the ability to
            # set events and add to the Channel's queue. The
            # function serves as a thread, so only start it once
            # and wait for the event to be set.
            LOG.debug("handling async data")
            if self.last_report == 0:
                LOG.debug("no data yet, waiting for new data...")
                self.get_data()
            else:
                LOG.debug('not fetching channel %s data',
                          self.name)
            LOG.debug("waiting on async event timeout")
            self.e_async.wait(self.async_timeout)
            # self.emit(time.time())
            self.try_emit()
            self.e_async.clear()

        elif self.mode == 'async2':
            # Used if the function defined doesn't return unless
            # a condition has been met. Simpler async case than
            # mode `async`.
            self.put_data(self.get_data())
            # self.emit(time.time())
            self.try_emit()

    def try_emit(self):
        """ Determine whether or not to run emit()

        If the report_rate duration has passed, or report_on_change is
        set to True, call emit()
        """
        t = time.time()
        _t = self.last_report + self.report_rate / 1000
        if self.report_on_change or t >= _t:
            self.emit(t)
        else:
            LOG.debug('not reporting channel %s',
                      self.name)

    def emit(self, timestamp):
        """ Set event to request data_in be sent.

        Sets Channel.e_emission event, which is watched by
        config_io.ChannelWatcher.run(). Sets last_report to the current
        timestamp.
        """
        self.e_emission.set()
        self.last_report = timestamp

    def _sleep(self):
        """ Sleep for the sample rate

        TODO: only sleep remaining time after data is processed.
        """
        time.sleep(self.sample_rate / 1000)

    def put_data(self, data):
        """ Place data in queue.

        In the future, this method will be switchable and optionally send
        data to gmq or SQL database, e.g. send_data_to_gmq(data)

        Parameters:
        data:           Datapoint to be placed in queue.
        """
        LOG.debug("putting: {}".format(data))
        self.q_out.put(Data(data, gain=self.multiplier, offset=self.offset))

    def run(self):
        """ Get data from source until Channel is stopped. """
        LOG.debug('starting')
        while not self.is_stopped():
            self.handle_getting_of_data()
        self.e_emission.set()
        LOG.debug('exiting')


class Data(tuple):
    """ Class to attach a timestamp to new data that is generated """
    def __new__(self, d, **kwargs):
        """ Subclasses tuple. (timestamp, data) """
        self.ts = time.time()
        self.offset = kwargs.get('offset')
        self.gain = kwargs.get('gain')
        try:
            self.d = d * self.gain + self.offset
        except TypeError:
            self.d = d
        # Attempt to JSON serialize the data point. If not
        # serializable, force into string.
        try:
            json.dumps(d)
        except TypeError:
            self.d = str(self.d)

        return tuple.__new__(Data, (self.ts, self.d))

    def age(self):
        """ Return age of data point in seconds """
        return time.time() - self.ts


class DownSampler(object):
    """ Class to deal with down sampling Channel data

    Methods:
    max:            Maximum value in list
    min:            Minimum value in list
    sum:            Sum of values in list
    avg:            Average of values in list
    act:            Last value in list ("actual value")
                    TODO: determine if this should return entire list
                          and if omitting down_sample in channel config
                          should result in the last value instead
    """
    method_mapper = {
        'max': max,
        'min': min,
        'sum': sum,
        'avg': lambda ls: float(sum(ls))/len(ls),
        'act': lambda ls: ls[-1]    # ls.pop()
    }

    def __init__(self, method):
        """
        Initialized by Channel.__init__()
        """
        self.method = method.lower()
        self._fn = self.method_mapper.get(self.method)

    def down_sample(self, data):
        """ Down sample data

        Assume data is list of tuples, e.g.
        [(ts, val), (ts, val), ...]
        """
        LOG.debug('Performing downsample %s on %s', self.method, data)
        try:
            return self._fn([d[1] for d in data])
        except:
            # kludge for non-tuple data
            return self._fn(data)



# pylint: disable=C0103,C0301,W1202
"""
Primary module to support edged tool.
"""
from __future__ import print_function
import os
import sys
import logging
import json
import time
import threading
from exoedge.exo_edge import ExoEdge
from exoedge.commands import Command as ExoSenseConnection
from exoedge.constants import OPTION_NAME_MAPPER, OPTION_TYPE_MAP, OPTION_PRECEDENCE, DEFAULTS

LOG = logging.getLogger('EXOEDGE.' + __name__)

class ExoCommand(ExoSenseConnection):
    """
    Create and start ExoSense client.

    usage:
        go [options]

    options:
        -h --help           Show this screen.
    """
    Name = 'go'

    def execute(self):
        from murano_client.client import MuranoClient
        from exoedge.options_handler import OptionsHandler


        # --------------------------------------
        # OPTIONS HANDLING
        # --------------------------------------
        # The value of an option comes from, in
        # decreasing priority:
        #   1. CLI switches
        #   2. Environment variables
        #   3. INI file
        #   4. Defaults
        # --------------------------------------
        OH_ARGS = {
            'cli': self.global_args,
            'env': os.environ,
            'dft': DEFAULTS
        }

        # -----------------------------------------
        # Include INI?
        _INI_FP = OH_ARGS['cli'].get('--ini-file') or OH_ARGS['env'].get('EDGED_INI_FILE') or None
        _NO_FS = OH_ARGS['cli'].get('--no-filesystem') or OH_ARGS['env'].get('EDGED_NO_FILESYSTEM') or None
        _INI_FILE = _INI_FP and not _NO_FS
        if _INI_FILE:
            from murano_client.atomicconfigparser import atomicconfigparser as IniParser
            I = IniParser(allow_no_value=True)
            I.read(_INI_FP)

            INI_VALUES = {}
            for s in I.sections():
                LOG.info('INI section %s...', s)
                for o, v in I.items(s):
                    LOG.info('INI option %s has value %s', o, v)
                    INI_VALUES.update({o: v})
            OH_ARGS.update({'ini': INI_VALUES})
        # -----------------------------------------


        OH = OptionsHandler(OPTION_PRECEDENCE,
                            OPTION_TYPE_MAP,
                            OPTION_NAME_MAPPER,
                            **OH_ARGS)
        PARAMETERS = OH.values
        LOG.info(json.dumps(PARAMETERS, indent=2))


        # -----------------------------------------
        # Define MuranoClient object
        # -----------------------------------------
        D = MuranoClient(**PARAMETERS)
        D.client_activate()
        TOKEN = D.murano_token() or PARAMETERS.get('murano_token') or None
        if _INI_FILE:
            if 'device' not in I.sections():
                I.add_section('device')
            I.set('device', 'murano_token', TOKEN)
            I.write(_INI_FP)
            LOG.critical('TOKEN saved to %s', _INI_FP)
        LOG.critical('TOKEN: %s', TOKEN)


        # -----------------------------------------
        # Define ExoEdge object
        # -----------------------------------------
        E = ExoEdge(D, **PARAMETERS)


        # -----------------------------------------
        # Start Reporting
        # -----------------------------------------
        D.start_client()
        E.setup()

        EDGED_TIMEOUT = self.global_args.get('--edged-timeout')
        def STOP_PROCESS():
            D.stop_all()
            E.stop()
            for thread in threading.enumerate():
                if hasattr(thread, 'stop'):
                    LOG.critical("Stopping thread: {}".format(thread))
                    if hasattr(thread, 'is_stopped'):
                        if not thread.is_stopped():
                            thread.stop()

            LOG.critical('TOKEN: %s', TOKEN)
            exit(0)

        try:
            if EDGED_TIMEOUT:
                EDGED_TIMEOUT = float(EDGED_TIMEOUT)
                BEGINNING = time.time()
                while time.time() < BEGINNING + EDGED_TIMEOUT:
                    time.sleep(1)
                STOP_PROCESS()
            else:
                while True:
                    time.sleep(1)
        except KeyboardInterrupt:
            STOP_PROCESS()
        return



# pylint: disable=W1202
import logging
from murano_client.client import StoppableThread

'''
What am I after:

I need a thread started once and only once.
 - Implemented by AsyncSource.get_async_source()
I need module-level functions to be callable via exoedge in the 'async' mode.
 - Implemented by AsyncSource.exoedge_put(data)
I need to be able to write functions so they are simple and developers can write their own.
 - ExoEdge Source developers can add class and/or module level functions/methods (or both) that implement the desired behavior.
I need a thread that can be extended to maintain a connection to a hardware resource.
 - ExoEdge Source developers can add members and methods to classes that subclass AsyncSource that manage hardware connections and interactions.
 - The AsyncSource class can be used in both "poll" and "async" config_io types.

'''

logging.getLogger('SOURCES')
logging.basicConfig(level=logging.DEBUG)

class AsyncSource(StoppableThread):
    """ Template class for defining custom "async" and/or "poll" type
    ExoEdge sources. Instances of this class will be a separate threads
    that follow the "borg" design pattern. """
    __borg_state = {}
    def __init__(self, **kwargs):
        """

        """
        self.__dict__ = self.__borg_state

        t_kwargs = {}
        if kwargs.get('group'):
            t_kwargs.update(group=kwargs.get('group'))
        if kwargs.get('target'):
            t_kwargs.update(target=kwargs.get('target'))
        if kwargs.get('name'):
            t_kwargs.update(name=kwargs.get('name'))
        if kwargs.get('args'):
            t_kwargs.update(args=kwargs.get('args'))
        if kwargs.get('kwargs'):
            t_kwargs.update(kwargs=kwargs.get('kwargs'))

        logging.info("async source thread kwargs: {}".format(t_kwargs))
        logging.info("async source other kwargs: {}".format(kwargs))

        super(AsyncSource, self).__init__(**t_kwargs)
        self.setDaemon(True)
        # parse the function pointer from ExoEdge for sending
        # data to ExoSense
        self._exoedge_put = kwargs.get('_fn')
        # parse the event object from ExoEdge for notifying
        # that there is new data for ExoSense
        self._exoedge_event = kwargs.get('event')

    def exoedge_put(self, data):
        """ Use this method to send async data to ExoSense. """

        self._exoedge_put(data)
        self._exoedge_event.set()

    def get_async_source(self):
        """ Call this function atleast once to start the
        Async Source thread. Call this function and assign
        to a variable to gain access to members and methods
        associated with this thread. """

        if not self.is_started():
            super(AsyncSource, self).start()
        return self



# pylint: disable=W0141,W0110,C0103
"""
ExoEdge ConfigIO module to handle Channel creation from a config_io object.
"""
from __future__ import print_function
import threading
import logging
import json
from exoedge.channel import Channel
from murano_client.client import StoppableThread

LOG = logging.getLogger('EXOEDGE.' + __name__)

class ConfigIO(StoppableThread, object):
    """
    This class provides methods to create channels from a
    config_io and to format the data_in object to be sent
    to Murano. ConfigIO must be supplied a config_io as a
    dictionary.
    """
    __lock = threading.Lock()
    e_send_data_in = threading.Event()
    e_new_config = threading.Event()

    def __init__(self,
                 config_io,
                 wait_timeout=1.0):
        StoppableThread.__init__(self, name="ConfigIO")
        self.config_io = {}
        self.wait_timeout = wait_timeout
        self._new_config_io = config_io
        self.last_report = 0
        self.channels = {}

    def stop(self):
        """ Stop all activity.

        Stops activity of all Channel threads, Channel-
        Watcher threads, and the ConfigIO thread.
        """
        self.stop_all_channels()
        super(ConfigIO, self).stop()

    def _has_channels(self):
        """ Return whether or not ConfigIO has channels.

        Used to determine if there are channels to stop.
        """
        return bool(self.channels)

    def _add_channel(self, name, ch_cfg):
        """ Create Channel and ChannelWatcher threads.

        Create a Channel and ChannelWatcher from a channel
        configuration object. Store them in the ConfigIO.channels
        dictionary.

        Parameters:
        ch_cfg:     the object representation of the channel
                    taken from config_io.
        name:       the key of the ch_cfg object in config_io

        """
        LOG.info('adding channel %s', name)
        ch_cfg['name'] = name
        channel = Channel(**ch_cfg)
        watcher = ChannelWatcher(channel,
                                 self.__lock,
                                 self.e_send_data_in)
        self.channels[name] = {'channel': channel,
                               'watcher': watcher}
        return

    def stop_all_channels(self):
        """ Stop Channel threads.

        Call channel.stop() for all Channels in ConfigIO.channels.
        Called by ConfigIO.run() when a new config_io is received.
        """
        LOG.info('stopping all channels')
        for name, d in self.channels.items():
            logging.debug('stopping channel %s', name)
            d['channel'].stop()
        self.channels = {}

    def set_config_io(self, cfg):
        """ Handle a new config_io.

        Handler for storing a newly received config_io and
        notifying the thread accordingly. Called by
        exo_edge.ConfigIOWatcher.run()

        Parameters:
        cfg:        the object representation of the config_io
        """
        LOG.critical('Received new config_io:\n%s',
                     json.dumps(cfg, indent=2))
        self._new_config_io = cfg
        self.e_new_config.set()

    def get_channel_data(self):
        """ Get data from Channel queues.

        Current implementation clears the queue after getting one value by
        either calling q.clear() or setting q to an empty list, per the
        implementation of LIFO queues.

        TODO: implement down_sampling
        """
        data_in = {}
        for name, d in self.channels.items():
            q = d['channel'].q_out
            if not q.empty():
                data_in[name] = d['channel'].downsampler.down_sample(
                    list(q.queue))
                # data_in[name] = q.get(timeout=0.0)
                try:
                    q.queue.clear()
                except AttributeError:  # LifoQueue.queue is a list
                    q.queue = []
            else:
                continue
        return data_in

    def run(self):
        """ Update ConfigIO and Channels upon new config_io.

        Waits for a new config_io event set by ConfigIO.set_config_io() and
        subsequently stops existing channels before creating new channels from
        the newly received config_io object. Starts Channels and ChannelWatchers.
        """
        LOG.debug('starting')
        self.e_new_config.set()     # set upon start()
        while not self.is_stopped():
            if self.e_new_config.wait(self.wait_timeout):
                LOG.debug('Received config_io')

                if self._has_channels():
                    self.stop_all_channels()

                if 'channels' in self._new_config_io.keys():
                    for name, cfg in self._new_config_io['channels'].items():
                        self._add_channel(name, cfg)

                    # start all Channels and ChannelWatchers
                    for channel_thread_group in self.channels.values():
                        for thread in channel_thread_group.values():
                            thread.start()

                else:
                    LOG.critical('Received new config_io without channels.')
                self.e_new_config.clear()
                self.config_io = self._new_config_io
            else:
                LOG.debug('no new config')
        LOG.debug('exiting')


class ChannelWatcher(StoppableThread, object):
    """ Watch for new data from Channel.

    One ChannelWatcher is created per Channel. It waits for new data from a
    Channel and notifies ConfigIO of this activity.
    """
    def __init__(self,
                 C,
                 lock,
                 request_event,
                 wait_timeout=1.0):
        """
        ChannelWatcher initialized by ConfigIO._add_channel()

        Parameters:
        C:              Channel instance
        lock:           threading.lock instance used to set ConfigIO's
                        request_event
        request_event:  event used to notify ConfigIO of activity. Currently
                        implemented as e_new_data_in
        """
        StoppableThread.__init__(self)
        self.C = C
        self.lock = lock
        self.e_send_data_in = request_event
        self.wait_timeout = wait_timeout
        self.name = 'watcher-{}'.format(C.name)

    def run(self):
        """ Wait for Channel emission and requests data_in be packaged.

        ChannelWatcher watches its associated Channel's e_emission event
        called by channel.Channel.emit().
        """
        LOG.debug('starting')

        while not self.C.is_stopped():
            if self.C.e_emission.wait(self.wait_timeout):
                LOG.debug('Received emission from Channel')
                with self.lock:
                    self.e_send_data_in.set()
                self.C.e_emission.clear()

        self.e_send_data_in.set()
        LOG.debug('exiting')




""" waves """
#from __future__ import print_function
import time
import math


def _apply_adsr(value,
    attack=3, # time in seconds
    decay=3, # time in seconds
    sustaint=3, # time in seconds
    sustainl=0.7, # level to sustain at (0.0-1.0)
    release=3, # time in seconds
    gap=3 # time in seconds
    ):
    def _mapper(x, in_min, in_max, out_min, out_max):
        return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
    stage = time.time() % (attack + decay + sustaint + release + gap)
    gain = 1
    if stage <= attack:
        # need to map range 0-1 over full attack time
        gain = _mapper(stage, 0, attack, 0, 1)
    elif stage <= (attack + decay):
        # need to map range 1-sustainl over full decay time
        gain = _mapper(stage, attack, attack+decay, 1, sustainl)
    elif stage <= (attack + decay + sustaint):
        gain = sustainl
    elif stage <= (attack + decay + sustaint + release):
        # need to map range sustainl-0 over full release time
        gain = _mapper(stage, attack + decay + sustaint, attack + decay + sustaint + release, sustainl, 0)
    elif stage <= (attack + decay + sustaint + release + gap):
        gain = 0

    return value * gain

def _wave_sine(r):
    return math.sin(2 * math.pi * r)

def _wave_cosin(r):
    return math.cos(2 * math.pi * r)

def _wave_square(r):
    if r < 0.5:
        return 1
    else:
        return 0

def _wave_triangle(ratio):
    if ratio == 0.5:
        return 1
    elif ratio < 0.5:
        return ratio * 2
    else:
        return ratio * 2 - 1
        
def _wave_saw_asc(ratio):
    return ratio

def _wave_saw_desc(ratio):
    return 1 - ratio

def _amp_lineargain(value, gain=1.0, offset=0.0):
    return (value * gain) + offset

def do_wave(shape='sin', period=60, amplifier=None, envolope=None, precision=2):
    """ Generate a wave from the current time

    Parameters:
    shape:          The shape of the wave
    period:         The period, in seconds, of the wave
    amplifier:      Amplifier to apply to the wave
    envolope:       ASDR envolope to apply to the wave
    precision:      Number of decimal points to return
    """

    ratio = (time.time() % period) / period

    if shape == 'sin':
        waved = _wave_sine(ratio)
    elif shape == 'cos':
        waved = _wave_cosin(ratio)
    elif shape == 'square':
        waved = _wave_square(ratio)
    elif shape == 'triangle':
        waved = _wave_triangle(ratio)
    elif shape == 'saw_asc':
        waved = _wave_saw_asc(ratio)
    elif shape == 'saw_desc':
        waved = _wave_saw_desc(ratio)
    else:
        waved = 1

    # Is there more than one kind of these?
    amped = waved
    if amplifier != None:
        if amplifier['type'] == 'lineargain':
            params = amplifier.copy()
            params['value'] = waved
            amped = _amp_lineargain(**params)

    if envolope != None:
        params = envolope.copy()
        params['value'] = amped
        envp = _apply_adsr(**params)
    else:
        envp = amped

    return round(envp, precision)


# def do_multi_wave(?array of waves?):
#     pass


# pylint: disable=W1202
"""
Namespace for ExoEdge commands.

TODO: Clear this out
"""

import sys
# import six
import docopt
import logging

LOG = logging.getLogger(name='GDC.HTTP')


class Command(object):  # ExositeConnection
    """Use methods of the HTTP(S) Device API.

  usage:
    http <command> [<args>...]

  commands:
    timestamp         Get the Murano timestamp.
    activate          Get a CIK/Token from a Murano Product for an id.
    read              Read resource described by <alias>.
    poll              Long poll <alias> for <request-timeout> since <if-modified-since>.
    write             Write <value> to <alias> resource.
    record            Record <value> to <alias> at time <timestamp>.
    content           Download content, get content info and list content.

  options:
    -N --no-ssl                 If provided, SSL/TLS will not be used in HTTP(S) comms.
    -F --no-form-encode         If provided, POST data will not be form-encoded.
    -P --no-product-domain      If provided, project domains (i.e. your project ID).
    -i --pid <id>               The Product ID.
    -t --timeout <secs>         If provided, <secs> is used as HTTP Timeout (in seconds).
                                Not implemented in Mqtt.
    -H --host <host>            If provided, <host> is used as the server hostname for
                                HTTP requests.
    -d --debug <lvl>            Turn on verbose debug output. Also logs curl commands.
    -u --uuid <uuid>            Specify the id of the connecting client (mqtt only).
    -f --file <file>            All operations will use <file> as the Device state file. This file
                                is compatible with Device class objects. Will store <cik> and
                                provision as the Murano client described in its options. This
                                option has special support for configuring the GWE config file
                                by using 'gwe' as the <file> argument.
    -C --cert <cert>            Use <cert> for Murano Provisioning of client described in
                                  the certificate subject.
    -K --pkey <pkey>            Use <pkey> in TLS communication with Murano.
    <cik>                       A CIK (Client Interface Key) or Token.
    <command>                   The HTTP(S) Device API method name.
    <args>                      Args for subcommands.

    """
    Name = 'http'
    def __init__(self, command_args, global_args):
        """
        Initialize the commands.
        :param command_args: arguments of the command
        :param global_args: arguments of the program
        """
        self.args = docopt.docopt(self.__doc__, argv=command_args, options_first=True)
        global_args.update({k:v for k, v in self.args.items() if v})
        self.global_args = global_args

        if self.global_args.get('--debug'):
            LOG.setLevel(eval('logging.'+self.global_args.get('--debug')))

    def execute(self, **exc_args):
        """Execute the commands"""
        import imp
        import os
        import pkgutil
        import murano_client.commands.http
        # Retrieve the command to execute.
        command_name = self.args.pop('<command>')
        # Retrieve the command arguments.
        command_args = self.args.pop('<args>')

        # Retrieve the module from the 'commands' package.
        sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'commands', Command.Name))
        try:
            http_path = murano_client.commands.http.__path__[0]
            LOG.debug('http_path: {}'.format(http_path))
            sys.path.insert(0, http_path)
            for _, module_name, _ in pkgutil.iter_modules([http_path]):
                command_path = os.path.join(http_path, command_name+'.py')
                if module_name == command_name:
                    LOG.debug('command_path: {}'.format(command_path))
                    the_module = imp.load_source(command_name, command_path)
            command_class = getattr(the_module, 'ExoCommand')

        except ImportError as exc:
            LOG.debug("{}: cannot find command {!r}: {}".format(self.Name, command_name, exc))
            raise docopt.DocoptExit()
        except AttributeError:
            LOG.debug('{}: unknown command: {}'.format(self.Name, command_name))
            raise docopt.DocoptExit()

        # Create an instance of the command and execute it
        command_class(command_args, self.global_args).execute()



"""
Installer for ExoSense Client
"""
import os
from setuptools import setup, find_packages
from exoedge import __version__

DOCS_URL = 'https://github.com/exosite/lib_exoedge_python'


def read(fname):
    """ Primarily used to open README file. """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

try:
    README = read('README.rst')
except:
    README = ''

setup(
    name="exoedge",
    version=__version__,
    author="Exosite LLC",
    author_email="support@exosite.com",
    description="""The ExoSense Client is the Python library for interacting with Exosite's ExoSense Industrial IoT Solution.""",
    license="Apache 2.0",
    keywords="murano exosite iot iiot gateway edge exoedge exosense",
    url="https://github.com/exosite/lib_exoedge_python",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'edged = exoedge.edged:main'
        ]
    },
    install_requires=['docopt>=0.6.2', 'murano-client>=18.7.1'],
    long_description=README,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Operating System Kernels :: Linux",
        "Topic :: Software Development :: Embedded Systems",
        "License :: OSI Approved :: Apache Software License",
    ],
    data_files=[])



"""
ExoEdge module. Provides high-level functionality to play nicely
with Exosite's Remote Condition Monitoring solution for Industrial IoT.
"""
from __future__ import print_function
import sys
import json
import logging
import threading
import time
from exoedge.config_io import ConfigIO
from murano_client.client import StoppableThread

PY_VERSION = sys.version_info[0]

LOG = logging.getLogger('EXOEDGE.' + __name__)


class ExoEdge(object):
    """
    ExoEdge is the root class for integration with ExoSense. An instance
    is created from a Device and is set up using one of several strategies:
        -   local:      Set up ExoEdge from a local file. Upon completion
                        write this value to Murano.
        -   remote:     Set up ExoEdge from config_io in Murano.
    TODO: GMQ support
    TODO: refactor to use q.put() and q.get()
    """

    def __init__(self,
                 D,
                 **kwargs):
        """ Onus of initialization is on the application.

        Parameters:
        D:                  Device/connection object from murano_client.
        strategy:           The strategy with which to instantiate an ExoSense
                            object.
                            local | read | subscribe
        config_io_file:     Local file used to cache copy of config_io.
                            Typically `./<UUID>.json`.
        cache_config_io:    Whether or not to save a local copy of
                            config_io in config_io_file.
        config_io_sync:     Whether or not to keep local copy of config_io
                            synced with Murano.
        """
        self.strategy = 'local' if kwargs.get('local_strategy') else 'remote'

        global config_io, device
        config_io = ConfigIO(None)
        device = D

        self.cache_config_io = not kwargs.get('no_config_cache')  # cache_config_io
        self.config_io_file = kwargs.get('config_io_file') or '{}.json'.format(kwargs.get('murano_id'))
        self.config_io_sync = not kwargs.get('no_config_sync')    # config_io_sync

        if self.config_io_sync:
            self.config_io_watcher = ConfigIOWatcher(self.config_io_file,
                                                     self.cache_config_io)

        self.data_in_writer = DataInWriter()
        self.__stop_event = threading.Event()

    def is_stopped(self):
        """ Check if the stop event is set."""
        return self.__stop_event.is_set()

    def stop(self):
        """ Stop reporting

        Calls config_io.ConfigIO.stop(), which eventually causes the Channels
        and their respective ChannelWatchers to stop gracefully.
        """
        LOG.info('stopping...')

        global config_io
        config_io.stop()

        if self.config_io_sync:
            self.config_io_watcher.stop()

        self.data_in_writer.stop()
        config_io.e_send_data_in.set()
        self.__stop_event.set()

    def setup(self):
        global config_io

        if self.strategy == 'local':
            config_io.set_config_io(self.read_local_config_io())
            self.tell_mur_config_io()

        elif self.strategy == 'remote':
            self.retrieve_remote_config_io()
            self.write_local_config_io()

        _ = self.config_io_watcher.start() if self.config_io_sync else None
        config_io.start()

        self.data_in_writer.start()

    def tell_mur_config_io(self):
        """ Write to Murano config_io resource. """
        device.tell(resource='config_io',
                    timestamp=time.time(),
                    payload=json.dumps(config_io._new_config_io))

    def write_local_config_io(self):
        """ Write config_io to local file.

        TODO: Rename method.
        """
        if self.cache_config_io:
            LOG.info('writing the following to local config io file %s:',
                     self.config_io_file)
            LOG.info(json.dumps(config_io._new_config_io, indent=2))
            with open(self.config_io_file, 'w') as f:
                json.dump(config_io._new_config_io, f)

    def read_local_config_io(self):
        """ Read config_io from local file. """
        infostr = 'READING LOCAL CONFIG_IO'
        LOG.critical('\n{:-^80}\n'.format(infostr))

        if self.config_io_file:
            try:
                return json.load(open(self.config_io_file, 'r'))

            except IOError:
                raise NoLocalConfigIO('config_io file {} does not exist.'
                                      .format(self.config_io_file))

        else:
            raise NoLocalConfigIO('A local config_io file has not been set.')

    def retrieve_remote_config_io(self):
        """ Get config_io from Murano """
        infostr = 'RETRIEVING CONFIG_IO FROM EXOSENSE'
        LOG.critical('\n{:-^80}\n'.format(infostr))

        inbound = device.watch(timeout=5.0)

        if inbound and inbound.resource == 'config_io':
            value = inbound.payload
            try:
                global config_io
                config_io.set_config_io(json.loads(value))
                self.tell_mur_config_io()

            except ValueError:
                LOG.critical('ERROR: `%s` is invalid JSON. config_io must be valid JSON', value)
        else:
            self.retrieve_remote_config_io()


class DataInWriter(StoppableThread, object):
    """ Class to deal with writing data_in.

    TODO: Rename class.
    TODO: refactor to use q.put() for murano_client
    """
    def __init__(self,
                 wait_timeout=1.0):
        """
        DataInWriter initialized by ExoEdge.go()

        Parameters:
        wait_timeout:       Timeout to wait for e_send_data_in event before
                            reevaluating own stop status
        """
        StoppableThread.__init__(self, name='DataInWriter')
        self.wait_timeout = wait_timeout

    def package_data_in(self, timeout=1.0):
        """ Assemble data_in JSON blob.

        Takes signal data and assembles into a packet that
        ExoEdge understands.

        TODO: Implement or remove timeout parameter.
        """
        data = config_io.get_channel_data()
        if data:
            return json.dumps(data)
        else:
            return None

    def write_data_in(self):
        """ Write to Murano resource data_in

        Called by run()
        """
        data = self.package_data_in()
        if data:
            LOG.critical('WRITING DATA: %s', data)
            device.tell(resource='data_in',
                        timestamp=time.time(),
                        payload=data)
        else:
            LOG.debug('NO DATA TO SEND')

    def run(self):
        """ Wait for e_send_data_in event to be set

        Event set by config_io.ChannelWatcher.run(). Upon set,
        write_data_in()
        """
        LOG.debug('starting')
        while not self.is_stopped():
            global config_io
            delivery = config_io.e_send_data_in.wait(self.wait_timeout)
            if delivery:
                self.write_data_in()

            config_io.e_send_data_in.clear()
        LOG.debug('exiting')


class ConfigIOWatcher(StoppableThread, object):
    """
    Class to deal with keeping local and cloud config_io
    in sync. Created if config_io_sync is set to True
    upon initiation of ExoEdge.

    TODO: refactor to q.get() from murano client
    """
    def __init__(self,
                 config_io_file=None,
                 cache_config_io=False,
                 timeout=300000):
        """
        Initialized by ExoEdge.go()

        Parameters:
        config_io_file:     File to save local copy of config_io
        cache_config_io:    Whether or not to save a local copy of config_io
        timeout:            Timeout to longpoll config_io resource in Murano
        """
        StoppableThread.__init__(self, name="ConfigIOWatcher")
        self.setDaemon(True)
        self.cache_config_io = cache_config_io
        self.config_io_file = config_io_file
        self.timeout = timeout

    def write_local_config_io(self):
        """ Write config_io to local file.

        TODO: Rename method.
        """
        LOG.debug('writing to local config io file %s',
                  self.config_io_file)
        with open(self.config_io_file, 'w') as f:
            json.dump(config_io._new_config_io, f)

    def run(self):
        """ Longpoll config_io.

        Informs config_io.ConfigIO that new config_io has been found.
        """
        while not self.is_stopped():
            global config_io, device
            inbound = device.watch(timeout=5)
            if inbound and inbound.resource == 'config_io':
                value = inbound.payload
                try:
                    if PY_VERSION < 3:
                        _json = json.loads(value)
                    elif PY_VERSION >= 3:
                        _json = json.load(value)
                    else:
                        LOG.critical("Version of Python is too old: %s", PY_VERSION)

                    config_io.set_config_io(_json)

                    device.tell(resource='config_io',
                                payload=inbound.payload,
                                timestamp=time.time())
                    _ = self.write_local_config_io() if self.cache_config_io else None

                except ValueError:
                    LOG.critical('ERROR: `%s` is invalid JSON. config_io must be valid JSON', value)

            else:
                LOG.debug('No config_io payload')


class ExoEdgeException(Exception):
    """
    Base exception class
    """
    pass


class ConfigIONotFound(ExoEdgeException):
    pass


class ConfigIONotSet(ConfigIONotFound):
    pass

class NoLocalConfigIO(ConfigIONotFound):
    pass

class MaxLongpollAttempts(ExoEdgeException):
    pass



import sys
import logging

LOG = logging.getLogger(name='EXOEDGE')
LOG_FORMAT = "%(asctime)s-%(levelname)s-%(name)s:%(funcName)s:%(lineno)d ::> %(message)s"
FORMATTER = logging.Formatter(LOG_FORMAT)
streamh = logging.StreamHandler(sys.stdout)
streamh.setFormatter(FORMATTER)
LOG.addHandler(streamh)
LOG.propagate = False


"""
Usage:
    edged [options] [<command>] [<args>...]

Commands:
    go          Start recording to ExoSense

Options:
    -h --help                       Show this screen.
    -v --version                    Print the current version of the ExoEdge library.
    -L --list-commands              Print installed and available commands.
    -i --ini-file <file>            INI file with device information.
    -c --config-io-file <cfgfile>   Local file in which to cache config_io. If
                                    `--local-strategy` flag is used, this file is
                                    expected to contain a valid config_io.
    -s --murano-id <sn>             The device serial number to use.
    -t --murano-token <token>       Token for device authentication.
    -K --pkeyfile <pkey>            Private key for TLS provisioning.
    -C --certfile <cert>            Public cert for TLS provisioning.
    -E --murano-cacert <cacert>     CA cert for PKI integration.
    -H --murano-host <host>         Set host for API requests.
    -p --murano-port <port>         Set port for API requests.
    -d --debug <lvl>                Tune the debug output. Logs curl commands at
                                    DEBUG.
                                    (DEBUG|INFO|WARNING|ERROR|CRITICAL).
    --local-strategy                Use `local` config_io strategy. If not present,
                                    use `remote` strategy.
    --no-filesystem                 Don't rely on a file system.
    --no-config-cache               Don't store a local copy of config_io.
    --no-config-sync                Don't keep config_io synced with ExoSense.
    --http-timeout <timeout>        Timeout to use for requests.
    --edged-timeout <timeout>       Timeout for edged process, in seconds.
    --watchlist <watchlist>         Murano resources to watch. Comma separated list.
                                        e.g. --watchlist=config_io,remote_control
    <command>                       The ExoEdge subcommand name.
    <args>                          Supported arguments for <command>

"""
from __future__ import print_function
import sys
import logging
import os
import imp
import pkgutil
from docopt import docopt, DocoptExit
import exoedge.commands
from exoedge import __version__ as VERSION
from exoedge.logger import LOG


def main():
    args = docopt(__doc__, version=VERSION, options_first=True)

    if args.get('--version'):
        print(VERSION)
        return

    if args.get('--list-commands'):
        sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname(__file__)), 'commands'))
        for _, module_name, _ in pkgutil.iter_modules(exoedge.commands.__path__):
            print(module_name)
        return

    if args.get('--debug'):
        LOG.setLevel(getattr(logging, args.get('--debug')) or 'CRITICAL')

    if args.get('<command>'):
        command_name = args.pop('<command>')
        argv = args.pop('<args>')
        if argv is None:
            argv = {}

        sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'commands'))
        try:

            exoedge_path = exoedge.__path__[0]
            LOG.debug('exoedge_path: %s', exoedge_path)

            _file, pathname, description = imp.find_module(command_name)
            LOG.debug('find_module: %s, %s, %s', _file, pathname, description)
            if _file:
                the_module = imp.load_module(command_name,
                                             _file,
                                             pathname,
                                             description)
            else:
                commands_path = exoedge.commands.__path__[0]
                LOG.debug('commands_path: %s', commands_path)
                for _, module_name, _ in pkgutil.iter_modules(
                        exoedge.commands.__path__):
                    if module_name == command_name:
                        command_path = os.path.join(commands_path,
                                                    command_name,
                                                    '__init__.py')
                        LOG.debug('command_path: %s', command_path)
                        the_module = imp.load_source(command_name,
                                                     command_path)
            LOG.debug('the_module: %s', the_module)
            command_class = getattr(the_module, 'ExoCommand')
        except ImportError as exc:
            raise DocoptExit("Cannot find command {!r}: {}"
                             .format(command_name, exc))

        command = command_class(argv, args)

        command.execute()
        return

    raise DocoptExit("Provide an option or subcommand\n")


if __name__ == '__main__':
    main()



"""
Module for storing option defaults and handling levels of
control
"""
from __future__ import print_function
import logging

LOG = logging.getLogger('EXOEDGE.' + __name__)


class TypeConverter(object):
    def __init__(self, _type):
        self._type = _type

    def convert(self, value):
        if value is None:
            return None
        else:
            return getattr(self, '_like_' + self._type.__name__)(value)

    def _like_str(self, v):
        if isinstance(v, str):
            return v
        else:
            return str(v)

    def _like_list(self, v):
        if isinstance(v, list):
            return v
        elif isinstance(v, str):
            return v.split(',')
        else:
            return [v]

    def _like_bool(self, v):
        if isinstance(v, bool):
            return v
        elif isinstance(v, str) and v in ['False', 'false', 'f', 'F', '0']:
            return False
        else:
            return bool(v)

    def _like_int(self, v):
        if isinstance(v, int):
            return v
        else:
            try:
                return int(v)
            except ValueError:
                return int()

    def _like_dict(self, v):
        if isinstance(v, dict):
            return v
        elif isinstance(v, str):
            try:
                return {_v.split(':')[0]:_v.split(':')[1] for _v in v.split(',')}
            except IndexError:
                return dict()


class OptionsHandler(object):
    def __init__(self,
                 precedence,
                 option_type_map,
                 option_name_map,
                 **kwargs):
        """
        Each value in kwargs is expected to be a dictionary.
        """
        self.precedence = [x for x in precedence if x in kwargs]
        self.option_type_map = option_type_map
        self.option_name_map = {k:v for k, v in option_name_map.items() if k in kwargs}
        self.opt_vals = kwargs

        self.values = self.honor_precedence()

    def _stack_params(self):
        param_stack = {}

        for _option, _type in self.option_type_map.items():
            option_stack = {}
            TC = TypeConverter(_type)

            for _option_source, _fn in self.option_name_map.items():
                _v = TC.convert(_fn(self.opt_vals.get(_option_source, None), _option))
                option_stack.update({_option_source: _v})

            param_stack.update({_option: option_stack})

        return param_stack

    def honor_precedence(self):
        param_stack = self._stack_params()
        _values = {}

        for option, stack in param_stack.items():
            _v = ([stack[_src] for _src in self.precedence if stack[_src] is not None] or [None])[0]
            _values.update({option: _v})

        return _values
